@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1997-2012. All Rights Reserved.
+%% Copyright Ericsson AB 1997-2018. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -68,7 +68,7 @@
 
 -type util_cpus() :: 'all' | integer() | [integer()].
 -type util_state() :: 'user' | 'nice_user' | 'kernel' | 'wait' | 'idle'.
--type util_value() :: {util_state(), float()} | float().
+-type util_value() :: [{util_state(), float()}] | float().
 -type util_desc() :: {util_cpus(), util_value(), util_value(), []}.
 
 %%----------------------------------------------------------------------
@@ -162,7 +162,8 @@ handle_call({?util, D, PC}, {Client, _Tag},
 	#state{os_type = {unix, Flavor}} = State) 
 	when Flavor == sunos;
 	     Flavor == linux;
-	     Flavor == freebsd ->
+	     Flavor == freebsd;
+	     Flavor == darwin ->
     case measurement_server_call(State#state.server, {?util, D, PC, Client}) of
 	{error, Reason} -> 
 	    {	reply, 
@@ -243,7 +244,7 @@ get_uint32_measurement(Request, #internal{os_type = {unix, Sys}}) when Sys == ir
     %% Get the load average using uptime.
     %% "8:01pm  up 2 days, 22:12,  4 users,  load average: 0.70, 0.58, 0.43"
     D = os:cmd("uptime") -- "\n",
-    Avg = lists:reverse(hd(string:tokens(lists:reverse(D), ":"))),
+    Avg = lists:reverse(hd(string:lexemes(lists:reverse(D), ":"))),
     {ok, [L1, L5, L15], _} = io_lib:fread("~f, ~f, ~f", Avg),
     case Request of
 	?avg1  -> sunify(L1);
@@ -531,11 +532,11 @@ measurement_server_loop(State) ->
 		    measurement_server_loop(State)
 	    end;
 	{Pid, Request} ->
-	    try get_uint32_measurement(Request, State) of
-		Result -> Pid ! {data, Result}
-	    catch
-		Error -> Pid ! {error, Error}
-	    end,
+            _ = try get_uint32_measurement(Request, State) of
+                    Result -> Pid ! {data, Result}
+                catch
+                    Error -> Pid ! {error, Error}
+                end,
 	    measurement_server_loop(State);
         {'EXIT', OldPid, _n} when State#internal.port == OldPid ->
 	    {ok, NewPid} = port_server_start_link(),
